﻿using System;
using System.Data.SqlClient;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
using System.Transactions;
using NUnit.Framework;

namespace NHibernate.Test.NHSpecificTest.NH3023 {
    [TestFixture]
    public class DeadlockConnectionPoolIssueTest : BugTestCase {
        protected override void OnSetUp() {
            RunScript("db-seed.sql");
        }
        protected override void OnTearDown() {
            RunScript("db-teardown.sql");
        }

        [Test]
        public void Connection_pool_may_become_corrupted_after_deadlock_in_distributed_transaction() {
            do {
                //
                // In order for this test to fail, there usually has to be an extreme load on the database.
                // I normally induce such a load by starting a trace of all Lock and Transaction events
                // in SQL Profiler.
                //

                var id = 1;

                try {
                    using(CreateDistributedTransactionScope()) {
                        using(var session = OpenSession())
                        using(session.BeginTransaction()) {

                            try {
                                new DeadlockHelper().ForceDeadlockOnConnection((SqlConnection)session.Connection);
                            } catch(SqlException x) {
                                //
                                // Deadlock error code is 1205.
                                //
                                if(x.Errors.Cast<SqlError>().Any(e => e.Number == 1205)) {
                                    //
                                    // It did what it was supposed to do ... THIS TIME
                                    //
                                    Console.WriteLine(x.Message);
                                    Console.WriteLine();
                                    continue;
                                }

                                //
                                // ? This shouldn't happen
                                //
                                Assert.Fail("Surprising exception when trying to force a deadlock: {0}", x.Message);
                            }

                            try {
                                session.Save(new DomainClass() {
                                    Id = id++,
                                    ByteData = new byte[] {1, 2, 3}
                                });

                                session.Flush();
                            } catch (Exception x) {
                                Console.WriteLine("Exception in initial session: {0}", x);
                                throw;
                            }
                        }
                    }
                }
                catch(AssertionException) {
                    throw;
                }
                catch(Exception x) {
                    //
                    // This probably shouldn't happen
                    //
                    Console.WriteLine("Initial session failed with error: {0}", x);
                    Console.WriteLine();
                }

                var subsequentFailedRequests = 0;

                for(var i = 1; i <= 10; i++) {
                    //
                    // The error message will vary on subsequent requests, so we'll somewhat 
                    // arbitrarily try 10
                    //

                    try {
                        using(var scope = CreateDistributedTransactionScope()) {
                            using(var session = OpenSession())
                            using(session.BeginTransaction()) {
                                try {
                                    session.Save(new DomainClass() {
                                        Id = id++,
                                        ByteData = new byte[] { 1, 2, 3 }
                                    });

                                    session.Flush();
                                }
                                catch(Exception x) {
                                    Console.WriteLine("Exception in subsequent session {0}: ", x);
                                    throw;
                                }
                            }

                            scope.Complete();
                        }
                    }
                    catch(Exception x) {
                        subsequentFailedRequests++;
                        Console.WriteLine("Subsequent session {0} failed with error: {1}", i, x);
                        Console.WriteLine();
                    }
                }

                //
                // Hopefully this will clean up the pool so that teardown can succeed
                //
                SqlConnection.ClearAllPools();

                Assert.Fail("Deadlock not reported on initial request; {0} subsequent requests failed.",
                    subsequentFailedRequests);

            } while(false);
            //
            // I'll change this to while(true) sometimes so I don't have to keep running the test
            //
        }

        private static TransactionScope CreateDistributedTransactionScope() {
            var scope = new TransactionScope();
            //
            // Forces promotion to distributed transaction
            //
            TransactionInterop.GetTransmitterPropagationToken(System.Transactions.Transaction.Current);
            return scope;
        }
        private void RunScript(string script) {
            var cxnString = cfg.Properties["connection.connection_string"] + "; Pooling=No";
                // Disable connection pooling so this won't be hindered by 
                // problems encountered during the actual test

            string sql;
            using(var reader = new StreamReader(GetType().Assembly.GetManifestResourceStream(GetType().Namespace + "." + script))) {
                sql = reader.ReadToEnd();
            }

            using(var cxn = new SqlConnection(cxnString)) {
                cxn.Open();

                foreach(var batch in Regex.Split(sql, @"^go\s*$", RegexOptions.IgnoreCase | RegexOptions.Multiline)
                    .Where(b => !string.IsNullOrEmpty(b))) {

                    using(var cmd = new System.Data.SqlClient.SqlCommand(batch, cxn)) {
                        cmd.ExecuteNonQuery();
                    }
                }

            }
        }
    }
}
