﻿using System;
using System.Data.SqlClient;
using System.Threading;
using System.Transactions;

namespace NHibernate.Test.NHSpecificTest.NH3023 {
    public class DeadlockHelper {
        public DeadlockHelper() {
            _victimHandle = new ManualResetEvent(false);
            _winnerHandle = new ManualResetEvent(false);
        }

        private readonly ManualResetEvent _victimHandle;
        private readonly ManualResetEvent _winnerHandle;

        public WaitHandle VictimHandle {
            get { return _victimHandle; }
        }
        public WaitHandle WinnerHandle {
            get { return _winnerHandle; }
        }

        public void ForceDeadlockOnConnection(SqlConnection connection) {
            //
            // Second thread with non-pooled connection, to deadlock
            // with current thread
            //
            var winnerThread = new Thread(() => {
                using(var scope = new TransactionScope(TransactionScopeOption.Required)) {
                    using(var cxn = new SqlConnection(connection.ConnectionString + ";Pooling=No")) {
                        cxn.Open();
                        DeadlockParticipant(cxn, false, _winnerHandle, _victimHandle);
                    }
                    scope.Complete();
                }
            });

            winnerThread.Start();

            try {
                //
                // This should always throw an exception of the form
                //  Transaction (Process ID nn) was deadlocked on lock resources with another process and has been chosen as the deadlock victim. Rerun the transaction.
                //
                DeadlockParticipant(connection, true, _victimHandle, _winnerHandle);
            }
            finally {
                winnerThread.Join();
                _winnerHandle.Reset();
                _victimHandle.Reset();
            }

            //
            // Should never get here
            //
            Console.WriteLine("Expected a deadlock exception, but it was not raised.");
        }
        private static void DeadlockParticipant(SqlConnection connection, bool isVictim, EventWaitHandle myHandle, WaitHandle partnerHandle) {
            //
            // CLID = 1 has only 10 records, CLID = 3 has 100.  This guarantees
            // which process will be chosen as the victim (the one which will be
            // less work to rollback)
            // 
            var clid = isVictim ? 1 : 3;
            using(var cmd = new System.Data.SqlClient.SqlCommand("UPDATE DeadlockHelper SET Data = newid() WHERE CLId = @CLID", connection)) {
                //
                // Exclusive lock on some records in the table
                //
                cmd.Parameters.AddWithValue("@CLID", clid);
                cmd.ExecuteNonQuery();
            }

            //
            // Notify partner that I have finished my work
            //
            myHandle.Set();
            //
            // Wait for partner to finish his work
            //
            partnerHandle.WaitOne();

            using(var cmd = new System.Data.SqlClient.SqlCommand("SELECT TOP 1 Data FROM DeadlockHelper ORDER BY Data", connection)) {
                //
                // Requires shared lock on table, should be blocked by
                // partner's exclusive lock
                //
                cmd.ExecuteNonQuery();
            }
        }
    }
}
